#include <linux/rt_lock.h>
#include <linux/sched.h>
#include <linux/init_task.h>
#include <linux/smp_lock.h>

#include "testtask.h"
#include "osfuncs.h"
#include "testfile.h"


/* #define DEBUG printf */
#define DEBUG(...)

testtask_t *tasks[MAX_TASKS];
struct semaphore *locks[MAX_LOCKS];
spinlock_t spinlocks[MAX_LOCKS];


void notrace setup_tasks()
{
  int i;
  for(i=0; i < nthreads; i++)
    {
      tasks[i] = create_testtask(i+1,priorities[i]);
    }
}

notrace void init_locks()
{
  int i;
  rt_init_MUTEX(&kernel_sem);
  locks[0] = &kernel_sem;
  for(i=1;i<MAX_LOCKS;i++)
    {
      locks[i] = calloc(1,sizeof(*locks[i]));
      rt_init_MUTEX(locks[i]);
      spinlocks[i] = SPIN_LOCK_UNLOCKED(spinlocks[i]);
    }
}



notrace void usage(char *arg0)
{
  printf("Usage: %s <testfile>\n",arg0);
  exit(1);
}

notrace void do_test(int taskid, test_t *test)
{
  testtask_t *t = tasks[taskid];
  switch(test->id)
    {
    case NOTEST:
      break;
    case BLOCKED:
      assert(t->blocked);
      break;
    case NONBLOCKED:
      assert(!t->blocked);
      break;
    case PRIORITY:
      assert(t->task->prio==test->data.priority);
      break;
    case STATE:
      assert(t->task->state==test->data.state);
      break;
    case LOCKCOUNT:
#ifdef CONFIG_DEBUG_PREEMPT
      assert(t->task->lock_count==test->data.lock_count);
#endif
      break;      
    }
}

int time = 0;
int current_taskid = -1;

timeentry_t *current_entry;

int notrace all_done(timeentry_t *entry)
{
  int i;
  int ndone = 0;
  for(i=0;i<nthreads;i++)
    {
      if(entry->jobs[i].done ||
	 tasks[i]->blocked ||
	 entry->jobs[i].id==IDLE)
	{
	  ndone++;
	}
    }
  return ndone == nthreads;
}
void notrace find_entry()
{
  if(all_done(current_entry) ||
     (current_entry->timeout && time > current_entry->timeout))
    {
      testentry_t *t;
      for(t=current_entry->tests; t; t=t->next)
	{
	  int i;
	  for(i=0; i<nthreads; i++)
	    {
	      do_test(i, &t->tests[i]);
	    }
	}
      time = 0;
      current_entry = current_entry->next;
      DEBUG("New entry\n");
      if(!current_entry)
	{
	  exit(0);
	}
    }
}

void notrace wakenext()
{
  int this_taskid = current_taskid;
 retry:
  find_entry();
  
#ifndef CONFIG_SMP
  if(irqflags || (current && preempt_count()))
    {
      return;
    }
#endif

  while(1)
    {
      current_taskid = (current_taskid+1) % nthreads;

      if(!tasks[current_taskid]->blocked && 
	 current_entry->jobs[current_taskid].id!=IDLE)
	{
	  if(current_taskid != this_taskid)
	    {
	      switch2testtask(tasks[current_taskid]);
	    }
	  return;
	}

      if(current_taskid == this_taskid)
	{
	  goto retry;
	}

    }
  
  return;
}


char *testfile;

void notrace betterErrorHandler(char *file, int line, char *statement)
{
  printf("%s:%d %s failed\n",file,line,statement); 
  printf("%s:%d was the last one executed\n",testfile,current_entry->line);
  abort(); 
}

int started = 0;

void notrace testtask_job()
{
  if(!started)
    {
      init_locks();
      errorHandler=betterErrorHandler;
      started = 1;
      
    }
  while(1)
    {
      timeentry_t *this_entry = current_entry;
      DEBUG("job: time: %d, current_taskid: %d done: %d\n", time, current_taskid, current_entry->done);
      time++;

      taskjob_t *job = &this_entry->jobs[current_taskid];
      if(!job->done)
	{
	  switch(job->id)
	    {
	    case IDLE:
	    case RUNNING:
	      break;
	    case TLOCK:
	      DEBUG("%d is locking %d\n",
		    current_taskid, job->data.lock);
	      if(job->data.lock==0) // BKL
		{
		  if(tasks[current_taskid]->task->lock_depth<0) 
		    {
		      down(locks[job->data.lock]);
		    }
		  tasks[current_taskid]->task->lock_depth++;
		}
	      else
		{
		  down(locks[job->data.lock]);
		}
		
	      DEBUG("%d has locked %d\n",
		     current_taskid, job->data.lock);
	      break;
	    case TLOCK_INTR:
	      DEBUG("%d is locking %d\n",
		    current_taskid, job->data.lock);
	      down_interruptible(locks[job->data.lock]);
	      DEBUG("%d has locked %d\n",
		     current_taskid, job->data.lock);
	      break;
	    case UNLOCK:
	      DEBUG("%d is unlocking %d\n",
		     current_taskid, job->data.lock);
	      if(job->data.lock==0) // BKL
		{
		  tasks[current_taskid]->task->lock_depth--;
		  if(tasks[current_taskid]->task->lock_depth<0) 
		    {
		      up(locks[job->data.lock]);
		    }
		}
	      else
		{
		  up(locks[job->data.lock]);
		}
	      assert(tasks[current_taskid]->task->state == TASK_RUNNING);
	      assert(!tasks[current_taskid]->task->blocked_on);
	      DEBUG("%d has unlocked %d\n",
		    current_taskid, job->data.lock);
	      break;
	    case SPIN_LOCK:
	      DEBUG("%d is spin-locking %d\n",
		    current_taskid, job->data.lock);
	      spin_lock(&spinlocks[job->data.lock]);
	      DEBUG("%d has spin-locked %d\n",
		     current_taskid, job->data.lock);
	      break;
	    case SPIN_UNLOCK:
	      DEBUG("%d is spin-unlocking %d\n",
		    current_taskid, job->data.lock);
	      spin_unlock(&spinlocks[job->data.lock]);
	      assert(!tasks[current_taskid]->task->blocked_on);
	      DEBUG("%d has spin-unlocked %d\n",
		     current_taskid, job->data.lock);
	      break;

	    case SIGNAL:
	      DEBUG("%d is signalling %d\n",
		    current_taskid, job->data.thread);
	      if(!(tasks[job->data.thread-1]->task->state & 
		   TASK_UNINTERRUPTIBLE)) 
		{
		  set_tsk_thread_flag(tasks[job->data.thread-1]->task, TIF_SIGPENDING);
		  wake_up_process(tasks[job->data.thread-1]->task);
		}
	      break;

	    case SETSTATE:
	      DEBUG("%d is setting it's state to %d\n",
		    current_taskid, job->data.state);
	      tasks[current_taskid]->task->state = job->data.state;
	      break;

	    default:
	      printf("Unkown job: %d\n",job->id);
	      exit(1);
	    }
	  assert(!irqflags);
	  assert(!preempt_count());

	  job->done = 1;
	}

      wakenext();
    }
}

int notrace main(int argc, char *argv[])
{
  if(argc!=2)
    {
      usage(argv[0]);
    }
  testfile = argv[1];
  read_testfile(testfile);
  current_entry = first_time;
  setup_tasks();
  sleep(1);
  wakenext();
  while(1)
    {
      sleep(1);
    }
  return 0;
}

void notrace __cyg_profile_func_enter (void *this_fn,
				void *call_site)
{
  if(!started)
    return;

/*   printf("call enter: time: %d, current_taskid: %d\n", time, current_taskid); */
  time+=10;
  wakenext();
}

void notrace __cyg_profile_func_exit  (void *this_fn,
			       void *call_site)
{
/*   printf("%s\n",__FUNCTION__); */
}

#ifdef CONFIG_SMP
void __lockfunc _raw_spin_lock(raw_spinlock_t *lock)
{
  int count = 0;
  while(!lock->raw_lock.slock)
    {
      time+=1;
      count++;
      assert(count<10000);
      wakenext();
    }
  lock->raw_lock.slock--;
  assert(lock->raw_lock.slock==0);
  lock->break_lock = current_taskid;
}


int __lockfunc _raw_spin_trylock(raw_spinlock_t *lock)
{
  if(!lock->raw_lock.slock)
    return 0;
  lock->raw_lock.slock--;
  assert(lock->raw_lock.slock==0);
  lock->break_lock = current_taskid;
  return 1;
}

unsigned long __lockfunc _raw_spin_lock_irqsave(raw_spinlock_t *lock)
{
  int count = 0;
  unsigned long res = tasks[current_taskid]->irqflag;
  tasks[current_taskid]->irqflag = 1;
  while(!lock->raw_lock.slock)
    {
      time+=1;
      count++;
      assert(count<10000);
      wakenext();
    }
  lock->raw_lock.slock--;
  assert(lock->raw_lock.slock==0);
  lock->break_lock = current_taskid;
  return res;
}

void __lockfunc _raw_spin_unlock(raw_spinlock_t *lock)
{
  assert( lock->break_lock == current_taskid );
  lock->raw_lock.slock++;
  assert(lock->raw_lock.slock==1);
}


void __lockfunc _raw_spin_unlock_irqrestore(raw_spinlock_t *lock, unsigned long flags)
{
  assert( lock->break_lock == current_taskid );
  tasks[current_taskid]->irqflag = flags;
  lock->raw_lock.slock++;
  assert(lock->raw_lock.slock==1);
}
#endif /* CONFIG_SMP */

void schedule() 
{ 
  assert(!irqflags);
  assert(!preempt_count());
  release_kernel_lock(current);

  if(!(___current->state==TASK_RUNNING))
    {
      tasks[current_taskid]->blocked = 1;
    }
  while(tasks[current_taskid]->blocked)
    {
      wakenext();
    }

  raw_local_irq_disable();
  reacquire_kernel_lock(current);
  raw_local_irq_enable();
}

int fastcall wake_up_process_common(struct task_struct * tsk, int state)
{
  int i;
  for(i=0; i < nthreads;i++)
    {
      if(tasks[i]->task==tsk)
	break;
    }
  assert(i<nthreads);
  tasks[i]->blocked = 0;
  tasks[i]->task->state = state;

  return 1;
}

int fastcall wake_up_process(struct task_struct * tsk)
{
  return wake_up_process_common(tsk, TASK_RUNNING);
}

int fastcall wake_up_process_mutex(struct task_struct * tsk)
{
  return wake_up_process_common(tsk, TASK_RUNNING_MUTEX);
}

void mutex_setprio(task_t *p, int prio) 
{ 
  p->prio = prio;
}

#if defined(CONFIG_DEBUG_PREEMPT) || defined(CONFIG_CRITICAL_TIMING)
void notrace add_preempt_count(unsigned int val)
{
  preempt_count() += val;
}


void notrace sub_preempt_count(unsigned int val)
{
  preempt_count() -= val;
}

void notrace add_preempt_count_ti(struct thread_info *ti, unsigned int val)
{
  preempt_count() += val;
}


void notrace sub_preempt_count_ti(struct thread_info *ti, unsigned int val)
{
  preempt_count() -= val;
}

void __might_sleep(char *file, int line)
{
}

#endif

#ifdef CONFIG_LATENCY_TRACE
void trace_irqs_off() 
{ 
  __raw_local_irq_disable(); 
}
void trace_irqs_on() 
{ 
  __raw_local_irq_enable(); 
}
#endif

DEFINE_RWLOCK(tasklist_lock);

void notrace defaultErrorHandler(char *file, int line, char *statement)
{
  printf("%s:%d %s failed\n",file,line,statement); 
  abort(); 
}

void (*errorHandler)(char *file, int line, char *statement) = 
defaultErrorHandler;
